/* Support for the generic parts of PE/PEI: the per-architecture dependent code
   Copyright 1995, 1996, 1997 Free Software Foundation, Inc.
   Written by Cygnus Support.

This file is part of BFD, the Binary File Descriptor library.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*
Most of this hacked by  Steve Chamberlain,
			sac@cygnus.com

PE/PEI rearrangement (and code added): Donn Terry
                                       Softway Systems, Inc.
*/

/* Hey look, some documentation [and in a place you expect to find it]!

   The main reference for the pei format is "Microsoft Portable Executable
   and Common Object File Format Specification 4.1".  Get it if you need to
   do some serious hacking on this code.

   Another reference:
   "Peering Inside the PE: A Tour of the Win32 Portable Executable
   File Format", MSJ 1994, Volume 9.

   The *sole* difference between the pe format and the pei format is that the
   latter has an MSDOS 2.0 .exe header on the front that prints the message
   "This app must be run under Windows." (or some such).
   (FIXME: Whether that statement is *really* true or not is unknown.
   Are there more subtle differences between pe and pei formats?
   For now assume there aren't.  If you find one, then for God sakes
   document it here!)

   The Microsoft docs use the word "image" instead of "executable" because
   the former can also refer to a DLL (shared library).  Confusion can arise
   because the `i' in `pei' also refers to "image".  The `pe' format can
   also create images (i.e. executables), it's just that to run on a win32
   system you need to use the pei format.

   FIXME: Please add more docs here so the next poor fool that has to hack
   on this code has a chance of getting something accomplished without
   wasting too much time.
*/

#include <libpei.h>

/**********************************************************************/

static void coff_swap_reloc_in PARAMS((bfd *, PTR, PTR));

static void
coff_swap_reloc_in (abfd, src, dst)
     bfd *abfd;
     PTR src;
     PTR dst;
{
  RELOC *reloc_src = (RELOC *) src;
  struct internal_reloc *reloc_dst = (struct internal_reloc *) dst;

  reloc_dst->r_vaddr = bfd_h_get_32(abfd, (bfd_byte *)reloc_src->r_vaddr);
  reloc_dst->r_symndx = bfd_h_get_signed_32(abfd, (bfd_byte *) reloc_src->r_symndx);

  reloc_dst->r_type = bfd_h_get_16(abfd, (bfd_byte *) reloc_src->r_type);
  reloc_dst->r_size = 0;

#ifdef IMAGE_REL_ALPHA_PAIR
  if (reloc_dst->r_type == IMAGE_REL_ALPHA_PAIR)
    {
      /* the MS compiler doesn't sign extend this 16 bit field, so we
	 do it here (it won't hurt if it already is, because it's only 16
	 bits) */
      reloc_dst->r_symndx = (reloc_dst->r_symndx<<16) >> 16;
    }
#endif

#ifdef ALPHA_DYN_R_COMB
  if (reloc_dst->r_type == ALPHA_DYN_R_COMB)
    {
      /* See coff-nt_alpha.c for details */
      reloc_dst->r_offset = reloc_dst->r_symndx >> 24;
      reloc_dst->r_symndx &= 0xffffff;
    }
#endif
}

static unsigned int
coff_swap_reloc_out (abfd, src, dst)
     bfd       *abfd;
     PTR	src;
     PTR	dst;
{
  struct internal_reloc *reloc_src = (struct internal_reloc *)src;
  struct external_reloc *reloc_dst = (struct external_reloc *)dst;
  long symndx;
#ifdef ALPHA_DYN_R_COMB
  if (reloc_src->r_type == ALPHA_DYN_R_COMB)
    {
      /* See coff-nt_alpha.c for details; r_symndx should never get too big,
	 and r_offset has already been range checked */
      /* Conditional because r_offset sometimes picks up junk */
      symndx = reloc_src->r_symndx | (reloc_src->r_offset << 24);
    }
  else
#endif
    symndx = reloc_src->r_symndx;
  bfd_h_put_32(abfd, reloc_src->r_vaddr, (bfd_byte *) reloc_dst->r_vaddr);
  bfd_h_put_32(abfd, symndx,             (bfd_byte *) reloc_dst->r_symndx);

  bfd_h_put_16(abfd, reloc_src->r_type, (bfd_byte *)
	       reloc_dst->r_type);

  return RELSZ;
}

static void coff_swap_filehdr_in PARAMS((bfd*, PTR, PTR));

static void
coff_swap_filehdr_in (abfd, src, dst)
     bfd            *abfd;
     PTR	     src;
     PTR	     dst;
{
  FILHDR *filehdr_src = (FILHDR *) src;
  struct internal_filehdr *filehdr_dst = (struct internal_filehdr *) dst;
  filehdr_dst->f_magic = bfd_h_get_16(abfd, (bfd_byte *) filehdr_src->f_magic);
  filehdr_dst->f_nscns = bfd_h_get_16(abfd, (bfd_byte *)filehdr_src-> f_nscns);
  filehdr_dst->f_timdat = bfd_h_get_32(abfd, (bfd_byte *)filehdr_src-> f_timdat);

  filehdr_dst->f_nsyms = bfd_h_get_32(abfd, (bfd_byte *)filehdr_src-> f_nsyms);
  filehdr_dst->f_flags = bfd_h_get_16(abfd, (bfd_byte *)filehdr_src-> f_flags);
  filehdr_dst->f_symptr = bfd_h_get_32 (abfd, (bfd_byte *) filehdr_src->f_symptr);
#ifdef COFF_IMAGE_WITH_PE
  /* There are really two magic numbers involved; the magic number that
     says this is a NT executable (PEI) and the magic number that determines
     the architecture.  If the NT magic number isn't valid, the architecture
     magic number could be mimiced by some other field (specifically,
     the number of relocs in section 3).  Since we know we're dealing with
     PEI here, we'll munge the magic number if it really isn't us. */

  if ( *(short *) &(filehdr_src->e_magic) != DOSMAGIC)
     filehdr_dst->f_magic = -1;
#endif
     
  /* Other people's tools sometimes generate headers
     with an nsyms but a zero symptr. */
  if (filehdr_dst->f_nsyms && filehdr_dst->f_symptr)
    {
      /* nothing */
    }
  else 
    {
      filehdr_dst->f_nsyms = 0;
    }

  filehdr_dst->f_opthdr = bfd_h_get_16(abfd, 
				       (bfd_byte *)filehdr_src-> f_opthdr);
}

#ifdef COFF_IMAGE_WITH_PE
#define coff_swap_filehdr_out pei_only_swap_filehdr_out
#else
#define coff_swap_filehdr_out pe_only_swap_filehdr_out
#endif

static void coff_swap_scnhdr_in PARAMS((bfd*, PTR, PTR));

static void
coff_swap_scnhdr_in (abfd, ext, in)
  bfd        *abfd;
  PTR	     ext;
  PTR	     in;
{
  SCNHDR *scnhdr_ext = (SCNHDR *) ext;
  struct internal_scnhdr *scnhdr_int = (struct internal_scnhdr *) in;

  memcpy(scnhdr_int->s_name, scnhdr_ext->s_name, sizeof(scnhdr_int->s_name));
  scnhdr_int->s_vaddr =
    GET_SCNHDR_VADDR (abfd, (bfd_byte *) scnhdr_ext->s_vaddr);
  scnhdr_int->s_paddr =
    GET_SCNHDR_PADDR (abfd, (bfd_byte *) scnhdr_ext->s_paddr);
  scnhdr_int->s_size =
    GET_SCNHDR_SIZE (abfd, (bfd_byte *) scnhdr_ext->s_size);
  scnhdr_int->s_scnptr =
    GET_SCNHDR_SCNPTR (abfd, (bfd_byte *) scnhdr_ext->s_scnptr);
  scnhdr_int->s_relptr =
    GET_SCNHDR_RELPTR (abfd, (bfd_byte *) scnhdr_ext->s_relptr);
  scnhdr_int->s_lnnoptr =
    GET_SCNHDR_LNNOPTR (abfd, (bfd_byte *) scnhdr_ext->s_lnnoptr);
  scnhdr_int->s_flags = bfd_h_get_32(abfd, (bfd_byte *) scnhdr_ext->s_flags);

  /* MS handles overflow of line numbers by carrying into the reloc field
     (it appears).  Since it's supposed to be zero for PE *IMAGE* format,
     that's safe.  This is still a bit iffy. */
#if defined(COFF_IMAGE_WITH_PE) && defined(DYNAMIC_LINKING)
  if ((abfd->file_flags & DYNAMIC) == 0 &&
     strcmp(scnhdr_int->s_name,".text") == 0)
    {
      scnhdr_int->s_nlnno = bfd_h_get_16(abfd, (bfd_byte *) scnhdr_ext->s_nlnno)
	    + (bfd_h_get_16(abfd, (bfd_byte *) scnhdr_ext->s_nreloc) << 16); 
      scnhdr_int->s_nreloc = 0;
    }
  else
    {
      scnhdr_int->s_nreloc = 
	  bfd_h_get_16(abfd, (bfd_byte *) scnhdr_ext->s_nreloc);
      scnhdr_int->s_nlnno = 
	  bfd_h_get_16(abfd, (bfd_byte *) scnhdr_ext->s_nlnno);
    }
#else
  scnhdr_int->s_nreloc = bfd_h_get_16(abfd, (bfd_byte *) scnhdr_ext->s_nreloc);
  scnhdr_int->s_nlnno = bfd_h_get_16(abfd, (bfd_byte *) scnhdr_ext->s_nlnno);
#endif

  /* If this is uninitialized data, use the virtual size (stored in
     s_paddr!!) instead of the physical size */
  if (scnhdr_int->s_flags & IMAGE_SCN_CNT_UNINITIALIZED_DATA)
    {
      scnhdr_int->s_size = scnhdr_int->s_paddr;
      scnhdr_int->s_paddr = 0;
    }
}

#define coff_mkobject pe_mkobject
static boolean
pe_mkobject (abfd)
     bfd * abfd;
{
  pe_data_type *pe;
  abfd->tdata.pe_obj_data = 
    (struct pe_tdata *) bfd_zalloc (abfd, sizeof (pe_data_type));

  if (abfd->tdata.pe_obj_data == 0)
    return false;

  pe = pe_data (abfd);

  pe->coff.pe = 1;

  /* in_reloc_p is architecture dependent */
  pe->in_reloc_p = in_reloc_p;
  return true;
}

#define coff_mkobject_hook pe_mkobject_hook

/* Create the COFF backend specific information.  */
static PTR
pe_mkobject_hook (abfd, filehdr, aouthdr)
     bfd * abfd;
     PTR filehdr;
     PTR aouthdr;
{
  struct internal_filehdr *internal_f = (struct internal_filehdr *) filehdr;
  pe_data_type *pe;

  if (!(internal_f->f_flags & F_DEBUG))
    abfd->file_flags |= HAS_DEBUG;

  if (pe_mkobject (abfd) == false)
    return NULL;

  pe = pe_data (abfd);
  pe->coff.sym_filepos = internal_f->f_symptr;
  /* These members communicate important constants about the symbol
     table to GDB's symbol-reading code.  These `constants'
     unfortunately vary among coff implementations...  */
  pe->coff.local_n_btmask = N_BTMASK;
  pe->coff.local_n_btshft = N_BTSHFT;
  pe->coff.local_n_tmask = N_TMASK;
  pe->coff.local_n_tshift = N_TSHIFT;
  pe->coff.local_symesz = SYMESZ;
  pe->coff.local_auxesz = AUXESZ;
  pe->coff.local_linesz = LINESZ;

  pe->coff.datestamp = internal_f->f_timdat;

  obj_raw_syment_count (abfd) =
    obj_conv_table_size (abfd) =
      internal_f->f_nsyms;

  pe->real_flags = internal_f->f_flags;

#ifdef COFF_IMAGE_WITH_PE
  if (aouthdr) 
    {
      pe->pe_opthdr = ((struct internal_aouthdr *)aouthdr)->pe;
    }
#ifdef DYNAMIC_LINKING
  if ((pe->pe_opthdr.DllCharacteristics & 0x0001) != 0)
    abfd->file_flags |= DYNAMIC;

  if ((pe->pe_opthdr.DllCharacteristics & 0x0002) != 0)
    abfd->file_flags |= NO_LINK;

#endif

#endif

  return (PTR) pe;
}

#ifdef COFF_IMAGE_WITH_PE
#define coff_bfd_copy_private_section_data pei_bfd_copy_private_section_data
#endif
